@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- encoding: utf-8 -*-
 
-# Copyright (c) 2002-2020 "Neo4j,"
+# Copyright (c) "Neo4j"
 # Neo4j Sweden AB [http://neo4j.com]
 #
 # This file is part of Neo4j.
@@ -19,29 +19,31 @@
 # limitations under the License.
 
 
-from collections import deque
 from logging import getLogger
 from random import random
-from time import perf_counter, sleep
-from warnings import warn
+from time import (
+    perf_counter,
+    sleep,
+)
 
+from neo4j.api import (
+    READ_ACCESS,
+    WRITE_ACCESS,
+)
 from neo4j.conf import SessionConfig
-from neo4j.api import READ_ACCESS, WRITE_ACCESS
-from neo4j.data import DataHydrator, DataDehydrator
+from neo4j.data import DataHydrator
 from neo4j.exceptions import (
+    ClientError,
+    IncompleteCommit,
     Neo4jError,
     ServiceUnavailable,
-    TransientError,
     SessionExpired,
+    TransientError,
     TransactionError,
-    ClientError,
 )
-from neo4j._exceptions import BoltIncompleteCommitError
 from neo4j.work import Workspace
-from neo4j.work.summary import ResultSummary
-from neo4j.work.transaction import Transaction
 from neo4j.work.result import Result
-from neo4j.io._bolt3 import Bolt3
+from neo4j.work.transaction import Transaction
 
 log = getLogger("neo4j")
 
@@ -113,11 +115,16 @@ class Session(Workspace):
             self._connection.send_all()
             self._connection.fetch_all()
             self._disconnect()
-        self._connection = self._pool.acquire(access_mode=access_mode, timeout=self._config.connection_acquisition_timeout, database=database)
+        self._connection = self._pool.acquire(
+            access_mode=access_mode,
+            timeout=self._config.connection_acquisition_timeout,
+            database=database,
+            bookmarks=self._bookmarks
+        )
 
     def _disconnect(self):
         if self._connection:
-            self._connection.in_use = False
+            self._pool.release(self._connection)
             self._connection = None
 
     def _collect_bookmark(self, bookmark):
@@ -130,6 +137,11 @@ class Session(Workspace):
             self._autoResult = None
             self._disconnect()
 
+    def _result_network_error(self):
+        if self._autoResult:
+            self._autoResult = None
+            self._disconnect()
+
     def close(self):
         """Close the session. This will release any borrowed resources, such as connections, and will roll back any outstanding transactions.
         """
@@ -213,8 +225,14 @@ class Session(Workspace):
 
         hydrant = DataHydrator()
 
-        self._autoResult = Result(cx, hydrant, self._config.fetch_size, self._result_closed)
-        self._autoResult._run(query, parameters, self._config.database, self._config.default_access_mode, self._bookmarks, **kwparameters)
+        self._autoResult = Result(
+            cx, hydrant, self._config.fetch_size, self._result_closed,
+            self._result_network_error
+        )
+        self._autoResult._run(
+            query, parameters, self._config.database,
+            self._config.default_access_mode, self._bookmarks, **kwparameters
+        )
 
         return self._autoResult
 
@@ -243,10 +261,21 @@ class Session(Workspace):
             self._transaction = None
             self._disconnect()
 
-    def _open_transaction(self, *, access_mode, database, metadata=None, timeout=None):
+    def _transaction_network_error_handler(self):
+        if self._transaction:
+            self._transaction = None
+            self._disconnect()
+
+    def _open_transaction(self, *, access_mode, database, metadata=None,
+                          timeout=None):
         self._connect(access_mode=access_mode, database=database)
-        self._transaction = Transaction(self._connection, self._config.fetch_size, self._transaction_closed_handler)
-        self._transaction._begin(database, self._bookmarks, access_mode, metadata, timeout)
+        self._transaction = Transaction(
+            self._connection, self._config.fetch_size,
+            self._transaction_closed_handler,
+            self._transaction_network_error_handler
+        )
+        self._transaction._begin(database, self._bookmarks, access_mode,
+                                 metadata, timeout)
 
     def begin_transaction(self, metadata=None, timeout=None):
         """ Begin a new unmanaged transaction. Creates a new :class:`.Transaction` within this session.
@@ -263,7 +292,7 @@ class Session(Workspace):
         :type metadata: dict
 
         :param timeout:
-            the transaction timeout in milliseconds.
+            the transaction timeout in seconds.
             Transactions that execute longer than the configured timeout will be terminated by the database.
             This functionality allows to limit query/transaction execution time.
             Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
@@ -308,10 +337,12 @@ class Session(Workspace):
                 try:
                     result = transaction_function(tx, *args, **kwargs)
                 except Exception:
-                    tx.rollback()
+                    tx.close()
                     raise
                 else:
                     tx.commit()
+            except IncompleteCommit:
+                raise
             except (ServiceUnavailable, SessionExpired) as error:
                 errors.append(error)
                 self._disconnect()
@@ -370,7 +401,7 @@ class Session(Workspace):
             with driver.session() as session:
                 values = session.read_transaction(get_two_tx)
 
-        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, \*args, \*\*kwargs)`
+        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, *args, **kwargs)`
         :param args: arguments for the `transaction_function`
         :param kwargs: key word arguments for the `transaction_function`
         :return: a result as returned by the given unit of work
@@ -395,7 +426,7 @@ class Session(Workspace):
                 node_id = session.write_transaction(create_node_tx, "example")
 
 
-        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, \*args, \*\*kwargs)`
+        :param transaction_function: a function that takes a transaction as an argument and does work with the transaction. `tx_function(tx, *args, **kwargs)`
         :param args: key word arguments for the `transaction_function`
         :param kwargs: key word arguments for the `transaction_function`
         :return: a result as returned by the given unit of work
@@ -442,7 +473,7 @@ def unit_of_work(metadata=None, timeout=None):
     :type metadata: dict
 
     :param timeout:
-        the transaction timeout in milliseconds.
+        the transaction timeout in seconds.
         Transactions that execute longer than the configured timeout will be terminated by the database.
         This functionality allows to limit query/transaction execution time.
         Specified timeout overrides the default timeout configured in the database using ``dbms.transaction.timeout`` setting.
